import logging # flake8: noqa
import collections as _collections

from vexmessage import Message as VexMessage

from vexbot.commands.restart_bot import restart_bot as _restart_bot
from vexbot.function_wrapers import (msg_list_wrapper,
                                     no_arguments)


class CommandManager:
    def __init__(self, messaging):
        # NOTE: commands is a dict of dicts and there is nested parsing
        self._commands = {}
        self._messaging = messaging
        self._commands['help'] = msg_list_wrapper(self._help)
        self._commands['commands'] = self._cmd_commands

    def register_command(self,
                         command: str,
                         function_or_dict: (_collections.Callable, dict)):
        """
        `func_or_nested` can either be a function or a dictionary
        """
        self._commands[command] = function_or_dict

    def is_command(self,
                   command: str,
                   call_command: bool=False) -> bool:

        callback, command, args = self._get_callback_recursively(command)
        if callback and call_command:
            callback(args)
            return True
        else:
            return bool(callback)

    def parse_commands(self, msg: VexMessage):
        command = msg.contents.get('command')

        if not command:
            return

        args = msg.contents.get('args')

        callback, command, args = self._get_callback_recursively(command, args)
        msg.contents['parsed_args'] = args

        if callback:
            results = callback(msg)
            # logging.error('{} {}'.format(results, command))

            if results:
                self._messaging.send_response(target=msg.source,
                                              original=command,
                                              response=results)

    def _get_callback_recursively(self,
                                  command: str,
                                  args: (list,
                                         str)=None) -> (_collections.Callable,
                                                        str,
                                                        list):
        """
        returns callback, command string, and args
        """

        if not command:
            return None, None, None
        if isinstance(args, str):
            args = args.split()
        elif args is None:
            args = []

        callback = self._commands.get(command)

        if isinstance(callback, _collections.Callable):
            return callback, command, args
        elif isinstance(callback, dict):
            dict_ = callback
        elif callback is None:
            return None, None, None
        else:
            s = '{} is not a callable function or a dict'
            raise TypeError(s.format(command))

        if not args:
            return None, None, None
        commands = []
        commands.append(command)

        # NOTE: Can't iterate None
        for command_number, command in enumerate(args):
            callback = dict_.get(command)
            if callback is None:
                return None, None, None

            commands.append(command)
            # dynamically reassigns the `dict_` value to travese nested dicts
            if isinstance(callback, dict):
                dict_ = callback
            elif isinstance(callback, _collections.Callable):
                break
            else:
                s = '{} is not a callable function or a dict'
                raise TypeError(s.format(callback.__name__))

        command_number += 1

        command_str = ' '.join(commands)
        return callback, command_str, args[command_number:]

    def _cmd_commands(self, msg: VexMessage):
        """
        returns a list of all available commands
        works recursively
        """
        def get_commands(d: dict):
            """
            recursive command
            """
            commands = []
            if not isinstance(d, dict):
                return commands

            for k, v in d.items():
                if isinstance(v, dict):
                    # NOTE: recursive command
                    stuff = get_commands(v)
                    for s in stuff:
                        commands.append(k + ' ' + s)
                else:
                    commands.append(k)
            return commands

        return get_commands(self._commands)

    def _send_command_not_found(self, target: str, original: str):
        """
        helper method
        """
        self._messaging.send_response(target=target,
                                      response='Command not found',
                                      original=original)

    def _help(self, args=None):
        if not args:
            return self._commands.keys()
        else:
            docs = []
            for arg in args:
                doc = self._commands.get(arg, None).__doc__
                if doc:
                    docs.append(doc)

            if docs:
                return docs


class BotCommandManager(CommandManager):
    def __init__(self, robot):
        super().__init__(robot.messaging)
        # nested command dict
        subprocess = {}

        # alias for pep8
        s_manager = robot.subprocess_manager

        # Store a reference to the subprocess settings for the `alive` command
        self._subprocess_settings = s_manager._settings
        subprocess['settings'] = msg_list_wrapper(s_manager.get_settings, 1)
        # subprocess['set-settings'] = msg_unpack_args(s_manager.update, 3)

        self._commands['subprocess'] = subprocess

        self._commands['killall'] = no_arguments(s_manager.killall)
        self._commands['restart_bot'] = no_arguments(_restart_bot)
        self._commands['alive'] = self._alive

        self._commands['start'] = msg_list_wrapper(s_manager.start)
        registered = s_manager.registered_subprocesses
        self._commands['subprocesses'] = no_arguments(registered)
        self._commands['restart'] = msg_list_wrapper(s_manager.restart)
        self._commands['kill'] = msg_list_wrapper(s_manager.kill)
        self._commands['terminate'] = msg_list_wrapper(s_manager.terminate)
        running = s_manager.running_subprocesses
        self._commands['running'] = no_arguments(running)

    def _alive(self, msg):
        """
        queries the subprocesses to check their status using messaging.

        can also use the `running` command to see what's running locally
        """
        # FIXME
        values = list(self._commands['subprocesses'](msg))

        process_names = []
        for value in values:
            setting = self._subprocess_settings.get(value)
            process_name = None
            if setting:
                if setting.get('--service_name'):
                    process_name = setting['--service_name']
            if process_name is None:
                process_name = value
            process_names.append(process_name)
        if process_names:
            try:
                process_names.remove(msg.source)
            except ValueError:
                pass
            for process_name in process_names:
                self._messaging.send_command(target=process_name,
                                             command='alive')


class AdapterCommandManager(CommandManager):
    def __init__(self, messaging):
        super().__init__(messaging)
        self._commands['alive'] = no_arguments(self._alive)

    def _alive(self, *args):
        self._messaging.send_status('CONNECTED')



import sys
import atexit
import signal
from subprocess import Popen, DEVNULL


class SubprocessManager:
    def __init__(self):
        # this is going to be a list of filepaths
        self._registered = {}
        self._settings = {}
        # these will be subprocesses
        self._subprocess = {}
        atexit.register(self._close_subprocesses)
        signal.signal(signal.SIGINT, self._handle_close_signal)
        signal.signal(signal.SIGTERM, self._handle_close_signal)
        self.blacklist = ['shell', ]

    def _handle_close_signal(self, signum=None, frame=None):
        self._close_subprocesses()
        sys.exit()

    def _close_subprocesses(self):
        """
        signum and frame are part of the signal lib
        """
        for process in self._subprocess.values():
            process.terminate()

    def registered_subprocesses(self):
        """
        returns all possible subprocesses that can be launched
        """
        return tuple(self._registered.keys())

    def register(self, key: str, value, settings: dict=None):
        self._settings[key] = settings
        if key in self.blacklist:
            return
        self._registered[key] = value

    def update_setting_value(self,
                             name: str,
                             setting_name: str,
                             setting_value):

        try:
            self._settings[name][setting_name] = setting_value
        except KeyError:
            pass

    def update_settings(self, name: str, settings: dict):
        """
        Need to pass in the subprocess name, a setting to be changed,
        and the value to change the setting to
        """
        old_settings = self._settings.get(name)
        if old_settings:
            old_settings.update(settings)
        else:
            self._settings[name] = settings

    def get_settings(self, key: str):
        """
        trys to get the settings information for a given subprocess. Passes
        silently when there is no information
        """
        settings = self._settings.get(key)
        return settings

    def start(self, keys: list):
        """
        starts subprocesses. Can pass in multiple subprocess to start
        """
        for key in keys:
            executable = self._registered.get(key)
            if executable is None:
                continue

            dict_list = [executable, ]
            settings = self._settings.get(key, {})
            filepath = settings.get('filepath')
            if filepath:
                dict_list.append(filepath)

            args = settings.get('args', [])
            if args:
                dict_list.extend(args)

            for k, v in settings.items():
                if k == 'filepath':
                    continue
                dict_list.append(k)
                dict_list.append(v)

            process = Popen(dict_list, stdout=DEVNULL)
            return_code = process.poll()
            if return_code is None:
                self._subprocess[key] = process

    def restart(self, values: list):
        """
        restarts subprocesses managed by the subprocess manager
        """
        for value in values:
            try:
                process = self._subprocess[value]
            except KeyError:
                continue

            process.terminate()
            self.start((value, ))

    def kill(self, values: list):
        """
        kills subprocess that is managed by the subprocess manager.
        If value does not match, quietly passes for now
        """
        for value in values:
            try:
                process = self._subprocess[value]
            except KeyError:
                continue
            process.kill()

    def terminate(self, values: list):
        for value in values:
            try:
                process = self._subprocess[value]
            except KeyError:
                continue
            process.terminate()

    def killall(self):
        """
        Kills every registered subprocess
        """
        for subprocess in self._subprocess.values():
            subprocess.kill()

    def running_subprocesses(self):
        """
        returns all running subprocess names
        """
        results = []
        killed = []
        for name, subprocess in self._subprocess.items():
            poll = subprocess.poll()
            if poll is None:
                results.append(name)
            else:
                killed.append(name)
        if killed:
            for killed_subprocess in killed:
                self._subprocess.pop(killed_subprocess)
        return results



import sys
import atexit
import signal
import asyncio
import argparse
import logging
import pkg_resources

import zmq
from zmq import ZMQError
from vexmessage import decode_vex_message

from vexbot.command_managers import AdapterCommandManager
from vexbot.adapters.messaging import ZmqMessaging

_IRC3_INSTALLED = True

try:
    pkg_resources.get_distribution('irc3')
except pkg_resources.DistributionNotFound:
    _IRC3_INSTALLED = False

if _IRC3_INSTALLED:
    import irc3
    from irc3.plugins.autojoins import AutoJoins

if _IRC3_INSTALLED:

    @irc3.plugin
    class AutoJoinMessage(AutoJoins):
        requires = ['irc3.plugins.core', ]

        def __init__(self, bot):
            super().__init__(bot)

        def connection_lost(self):
            self.bot.messaging.send_status('DISCONNECTED')
            super(AutoJoinMessage, self).connection_lost()

        def join(self, channel=None):
            super(AutoJoinMessage, self).join(channel)
            self.bot.messaging.send_status('CONNECTED')

        @irc3.event(irc3.rfc.KICK)
        def on_kick(self, mask, channel, target, **kwargs):
            self.bot.messaging.send_status('DISCONNECTED')
            super().on_kick(mask, channel, target, **kwargs)

        @irc3.event("^:\S+ 47[1234567] \S+ (?P<channel>\S+).*")
        def on_err_join(self, channel, **kwargs):
            self.bot.messaging.send_status('DISCONNECTED')
            super().on_err_join(channel, **kwargs)

    @irc3.plugin
    class EchoToMessage(object):
        requires = ['irc3.plugins.core',
                    'irc3.plugins.command']

        def __init__(self, bot):
            self.bot = bot

        @irc3.event(irc3.rfc.PRIVMSG)
        def message(self, mask, event, target, data):
            nick = mask.nick
            message = data
            self.bot.messaging.send_message(author=nick,
                                            message=str(message),
                                            channel=target)


def create_irc_bot(nick,
                   password,
                   host=None,
                   port=6667,
                   realname=None,
                   channel=None):

    config = dict(ssl=False,
                  includes=['irc3.plugins.core',
                            'irc3.plugins.command',
                            'irc3.plugins.human',
                            __name__])

    if realname is None:
        realname = nick
    if channel is None:
        channel = nick

    config['nick'] = nick
    config['password'] = password
    config['host'] = host
    config['port'] = port
    config['username'] = realname
    config['autojoins'] = channel
    config['level'] = 30

    bot = irc3.IrcBot.from_config(config)

    return bot


async def _check_subscription(bot):
    while True:
        await asyncio.sleep(1)
        msg = None
        try:
            msg = bot.messaging.sub_socket.recv_multipart(zmq.NOBLOCK)
        except zmq.error.Again:
            pass
        if msg:
            msg = decode_vex_message(msg)
            if msg.type == 'CMD':
                bot.command_parser.parse_commands(msg)
            elif msg.type == 'RSP':
                channel = msg.contents.get('channel')
                message = msg.contents.get('response')
                bot.privmsg(channel, message)
                bot.messaging.send_message(author=bot.nick,
                                           message=message,
                                           channel=channel)


def _default(*args, **kwargs):
    pass


def _call_func_with_arg(func, arg):
    def inner(*args):
        return func(arg)
    return inner


def _send_disconnected(messaging):
    def inner():
        messaging.send_status('DISCONNECTED')
    return inner


def _handle_close(messaging, event_loop):
    def inner(signum=None, frame=None):
        _send_disconnected(messaging)()
        event_loop.stop()
        for task in asyncio.Task.all_tasks():
            task.cancel()
    return inner


def main(nick,
         password,
         host,
         channel,
         publish_address,
         subscribe_address,
         service_name):

    if not _IRC3_INSTALLED:
        logging.error('irc requires `irc3` to be installed. Please install '
                      'using `pip install irc3`')

    irc_client = create_irc_bot(nick,
                                password,
                                host,
                                channel=channel)

    try:
        messaging = ZmqMessaging(service_name,
                                 publish_address,
                                 subscribe_address,
                                 socket_filter=service_name)

        messaging.start_messaging()
    except ZMQError:
        return
    # Duck type messaging onto irc_client, FTW
    irc_client.messaging = messaging
    command_parser = AdapterCommandManager(messaging)
    irc_client.command_parser = command_parser

    """
    command_parser.register_command('server config', _default)
    command_parser.register_command('ip', _default)
    command_parser.register_command('join', _default)
    command_parser.register_command('kick', _default)
    command_parser.register_command('part', _default)
    command_parser.register_command('invite', _default)
    command_parser.register_command('topic', _default)
    command_parser.register_command('away', _default)
    """

    irc_client.create_connection()
    irc_client.add_signal_handlers()
    event_loop = asyncio.get_event_loop()
    asyncio.ensure_future(_check_subscription(irc_client))
    atexit.register(_send_disconnected(messaging))

    handle_close = _handle_close(messaging, event_loop)
    signal.signal(signal.SIGINT, handle_close)
    signal.signal(signal.SIGTERM, handle_close)
    try:
        event_loop.run_forever()
    except KeyboardInterrupt:
        pass
    event_loop.close()
    sys.exit()


def _get_args():
    parser = argparse.ArgumentParser()
    parser.add_argument('--nick')
    parser.add_argument('--password')
    parser.add_argument('--host')
    parser.add_argument('--channel')
    parser.add_argument('--publish_address')
    parser.add_argument('--subscribe_address')
    parser.add_argument('--service_name')

    return parser.parse_args()


if __name__ == '__main__':

    kwargs = vars(_get_args())
    main(**kwargs)
